@@ -6,7 +6,7 @@
 #     [  Github: https://github.com/Breakthrough/PySceneDetect/  ]
 #     [  Documentation: http://pyscenedetect.readthedocs.org/    ]
 #
-# Copyright (C) 2014-2020 Brandon Castellano <http://www.bcastell.com>.
+# Copyright (C) 2014-2021 Brandon Castellano <http://www.bcastell.com>.
 #
 # PySceneDetect is licensed under the BSD 3-Clause License; see the included
 # LICENSE file, or visit one of the following pages for details:
@@ -42,8 +42,6 @@ module and the scenedetect.cli.CliContext object.
 
 # Standard Library Imports
 from __future__ import print_function
-import sys
-import string
 import logging
 
 # Third-Party Library Imports
@@ -51,15 +49,14 @@ import click
 
 # PySceneDetect Library Imports
 import scenedetect
-from scenedetect.cli.context import CliContext
-from scenedetect.frame_timecode import FrameTimecode
-from scenedetect.video_manager import VideoManager
 
-from scenedetect.video_splitter import is_mkvmerge_available
-from scenedetect.video_splitter import is_ffmpeg_available
+from scenedetect.cli.context import check_split_video_requirements
+from scenedetect.cli.context import contains_sequence_or_url
+from scenedetect.cli.context import parse_timecode
 
 from scenedetect.platform import get_and_create_path
-
+from scenedetect.platform import init_logger
+logger = logging.getLogger('pyscenedetect')
 
 def get_help_command_preface(command_name='scenedetect'):
     """ Preface/intro help message shown at the beginning of the help command. """
@@ -104,29 +101,6 @@ def add_cli_command(cli, command):
     COMMAND_DICT.append(command)
 
 
-def parse_timecode(cli_ctx, value):
-    # type: (CliContext, str) -> Union[FrameTimecode, None]
-    """ Parses a user input string expected to be a timecode, given a CLI context.
-
-    Returns:
-        (FrameTimecode) Timecode set to value with the CliContext VideoManager framerate.
-            If value is None, skips processing and returns None.
-
-    Raises:
-        click.BadParameter
-     """
-    cli_ctx.check_input_open()
-    if value is None:
-        return value
-    try:
-        timecode = FrameTimecode(
-            timecode=value, fps=cli_ctx.video_manager.get_framerate())
-        return timecode
-    except (ValueError, TypeError):
-        raise click.BadParameter(
-            'timecode must be in frames (1234), seconds (123.4s), or HH:MM:SS (00:02:03.400)')
-
-
 def print_command_help(ctx, command):
     # type: (click.Context, Callable[]) -> None
     """ Print Command Help: Prints PySceneDetect help/usage for a given command. """
@@ -170,7 +144,7 @@ def duplicate_command(ctx, param_hint):
     error_strs.append('Error: Command %s specified multiple times.' % param_hint)
     error_strs.append('The %s command may appear only one time.')
 
-    logging.error('\n'.join(error_strs))
+    ctx.obj.logger.error('\n'.join(error_strs))
     raise click.BadParameter('\n  Command %s may only be specified once.' % param_hint,
                              param_hint='%s command' % param_hint)
 
@@ -181,9 +155,10 @@ def duplicate_command(ctx, param_hint):
 @click.option(
     '--input', '-i',
     multiple=True, required=False, metavar='VIDEO',
-    type=click.Path(exists=True, file_okay=True, readable=True, resolve_path=True), help=
+    type=click.STRING, help=
     '[Required] Input video file.'
-    ' May be specified multiple times to concatenate several videos together.')
+    ' May be specified multiple times to concatenate several videos together.'
+    ' Also supports image sequences and URLs.')
 @click.option(
     '--output', '-o',
     multiple=False, required=False, metavar='DIR',
@@ -208,6 +183,15 @@ def duplicate_command(ctx, param_hint):
     'Skips N frames during processing (-fs 1 skips every other frame, processing 50% of the video,'
     ' -fs 2 processes 33% of the frames, -fs 3 processes 25%, etc...).'
     ' Reduces processing speed at expense of accuracy.')
+@click.option(
+    '--min-scene-len', '-m', metavar='TIMECODE',
+    type=click.STRING, default='0.6s', show_default=True, help=
+    'Minimum size/length of any scene. TIMECODE can be specified as exact'
+    ' number of frames, a time in seconds followed by s, or a timecode in the'
+    ' format HH:MM:SS or HH:MM:SS.nnn')
+@click.option(
+    '--drop-short-scenes', is_flag=True, flag_value=True, help=
+    'Drop scenes shorter than `--min-scene-len` instead of combining them with neighbors')
 @click.option(
     '--stats', '-s', metavar='CSV',
     type=click.Path(exists=False, file_okay=True, writable=True, resolve_path=False), help=
@@ -229,12 +213,12 @@ def duplicate_command(ctx, param_hint):
 @click.option(
     '--quiet', '-q',
     is_flag=True, flag_value=True, help=
-    'Suppresses all output of PySceneDetect except for those from the specified'
-    ' commands. Equivalent to setting `--verbosity none`. Overrides the current verbosity'
-    ' level, even if `-v`/`--verbosity` is set.')
+    'Suppresses all output of PySceneDetect to the terminal/stdout. If a logfile is'
+    ' specified, it will still be generated with the specified verbosity.')
 @click.pass_context
 # pylint: disable=redefined-builtin
-def scenedetect_cli(ctx, input, output, framerate, downscale, frame_skip, stats,
+def scenedetect_cli(ctx, input, output, framerate, downscale, frame_skip,
+                    min_scene_len, drop_short_scenes, stats,
                     verbosity, logfile, quiet):
     """ For example:
 
@@ -250,49 +234,33 @@ def scenedetect_cli(ctx, input, output, framerate, downscale, frame_skip, stats,
 
     logging.disable(logging.NOTSET)
 
-    format_str = '[PySceneDetect] %(message)s'
-    if verbosity.lower() == 'none':
-        verbosity = None
-    elif verbosity.lower() == 'debug':
-        format_str = '%(levelname)s: %(module)s.%(funcName)s(): %(message)s'
-
-    if quiet:
-        verbosity = None
+    verbosity = getattr(logging, verbosity.upper()) if verbosity is not None else None
+    init_logger(log_level=verbosity, show_stdout=not quiet, log_file=logfile)
 
+    ctx.obj.quiet_mode = True if quiet else False
     ctx.obj.output_directory = output
-    if logfile is not None:
-        logfile = get_and_create_path(logfile)
-        logging.basicConfig(
-            filename=logfile, filemode='a', format=format_str,
-            level=getattr(logging, verbosity.upper()) if verbosity is not None else verbosity)
-        logging.info('Version: %s', scenedetect.__version__)
-        logging.info('Info Level: %s', verbosity)
-    else:
-        if verbosity is not None:
-            logging.basicConfig(format=format_str,
-                                level=getattr(logging, verbosity.upper()))
-        else:
-            logging.disable(logging.CRITICAL)
 
-    ctx.obj.quiet_mode = True if verbosity is None else False
+    ctx.obj.logger.info('PySceneDetect %s', scenedetect.__version__)
 
     if stats is not None and frame_skip != 0:
         ctx.obj.options_processed = False
         error_strs = [
             'Unable to detect scenes with stats file if frame skip is not 1.',
             '  Either remove the -fs/--frame-skip option, or the -s/--stats file.\n']
-        logging.error('\n'.join(error_strs))
+        ctx.obj.logger.error('\n'.join(error_strs))
         raise click.BadParameter(
             '\n  Combining the -s/--stats and -fs/--frame-skip options is not supported.',
             param_hint='frame skip + stats file')
+
     try:
         if ctx.obj.output_directory is not None:
-            logging.info('Output directory set:\n  %s', ctx.obj.output_directory)
+            ctx.obj.logger.info('Output directory set:\n  %s', ctx.obj.output_directory)
         ctx.obj.parse_options(
             input_list=input, framerate=framerate, stats_file=stats, downscale=downscale,
-            frame_skip=frame_skip)
-    except:
-        logging.error('Could not parse CLI options.')
+            frame_skip=frame_skip, min_scene_len=min_scene_len, drop_short_scenes=drop_short_scenes)
+
+    except Exception as ex:
+        ctx.obj.logger.error('Could not parse CLI options.: %s', ex)
         raise
 
 
@@ -411,13 +379,11 @@ def time_command(ctx, start, duration, end):
     'Threshold value (float) that the content_val frame metric must exceed to trigger a new scene.'
     ' Refers to frame metric content_val in stats file.')
 @click.option(
-    '--min-scene-len', '-m', metavar='TIMECODE',
-    type=click.STRING, default="0.6s", show_default=True, help=
-    'Minimum size/length of any scene. TIMECODE can be specified as exact'
-    ' number of frames, a time in seconds followed by s, or a timecode in the'
-    ' format HH:MM:SS or HH:MM:SS.nnn')
+    '--luma-only', '-l',
+    is_flag=True, flag_value=True, help=
+    'Only consider luma/brightness channel (useful for greyscale videos).')
 @click.pass_context
-def detect_content_command(ctx, threshold, min_scene_len):
+def detect_content_command(ctx, threshold, luma_only):
     """ Perform content detection algorithm on input video(s).
 
     detect-content
@@ -425,17 +391,72 @@ def detect_content_command(ctx, threshold, min_scene_len):
     detect-content --threshold 27.5
     """
 
-    min_scene_len = parse_timecode(ctx.obj, min_scene_len)
-
-    logging.debug('Detecting content, parameters:\n'
-                  '  threshold: %d, min-scene-len: %d',
-                  threshold, min_scene_len)
+    min_scene_len = 0 if ctx.obj.drop_short_scenes else ctx.obj.min_scene_len
+    luma_mode_str = '' if not luma_only else ', luma_only mode'
+    ctx.obj.logger.debug('Detecting content, parameters:\n'
+                  '  threshold: %d, min-scene-len: %d%s',
+                  threshold, min_scene_len, luma_mode_str)
 
     # Initialize detector and add to scene manager.
     # Need to ensure that a detector is not added twice, or will cause
     # a frame metric key error when registering the detector.
     ctx.obj.add_detector(scenedetect.detectors.ContentDetector(
-        threshold=threshold, min_scene_len=min_scene_len))
+        threshold=threshold, min_scene_len=min_scene_len, luma_only=luma_only))
+
+
+@click.command('detect-adaptive')
+@click.option(
+    '--threshold', '-t', metavar='VAL',
+    type=click.FLOAT, default=3.0, show_default=True, help=
+    'Threshold value (float) that the calculated frame score must exceed to'
+    ' trigger a new scene (see frame metric adaptive_ratio in stats file).')
+@click.option(
+    '--min-scene-len', '-m', metavar='TIMECODE',
+    type=click.STRING, default="0.6s", show_default=True, help=
+    'Minimum size/length of any scene. TIMECODE can be specified as exact'
+    ' number of frames, a time in seconds followed by s, or a timecode in the'
+    ' format HH:MM:SS or HH:MM:SS.nnn')
+@click.option(
+    '--min-delta-hsv', '-d', metavar='VAL',
+    type=click.FLOAT, default=15.0, show_default=True, help=
+    'Minimum threshold (float) that the content_val must exceed in order to register as a new'
+    ' scene. This is calculated the same way that `detect-content` calculates frame score.')
+@click.option(
+    '--frame-window', '-w', metavar='VAL',
+    type=click.INT, default=2, show_default=True, help=
+    'Size of window (number of frames) before and after each frame to average together in'
+    ' order to detect deviations from the mean.')
+@click.option(
+    '--luma-only', '-l',
+    is_flag=True, flag_value=True, help=
+    'Only consider luma/brightness channel (useful for greyscale videos).')
+@click.pass_context
+def detect_adaptive_command(ctx, threshold, min_scene_len, min_delta_hsv,
+                            frame_window, luma_only):
+    """ Perform adaptive detection algorithm on input video(s).
+
+    detect-adaptive
+
+    detect-adaptive --threshold 3.2
+    """
+
+    min_scene_len = parse_timecode(ctx.obj, min_scene_len)
+    luma_mode_str = '' if not luma_only else ', luma_only mode'
+
+    ctx.obj.logger.debug('Adaptively detecting content, parameters:\n'
+                  '  threshold: %d, min-scene-len: %d%s',
+                  threshold, min_scene_len, luma_mode_str)
+
+    # Initialize detector and add to scene manager.
+    # Need to ensure that a detector is not added twice, or will cause
+    # a frame metric key error when registering the detector.
+    ctx.obj.add_detector(scenedetect.detectors.AdaptiveDetector(
+        video_manager=ctx.obj.video_manager,
+        adaptive_threshold=threshold,
+        min_scene_len=min_scene_len,
+        min_delta_hsv=min_delta_hsv,
+        window_width=frame_window,
+        luma_only=luma_only))
 
 
 
@@ -445,12 +466,6 @@ def detect_content_command(ctx, threshold, min_scene_len):
     type=click.IntRange(0, 255), default=12, show_default=True, help=
     'Threshold value (integer) that the delta_rgb frame metric must exceed to trigger a new scene.'
     ' Refers to frame metric delta_rgb in stats file.')
-@click.option(
-    '--min-scene-len', '-m', metavar='TIMECODE',
-    type=click.STRING, show_default=True, default="0.6s", help=
-    'Minimum size/length of any scene. TIMECODE can be specified as exact'
-    ' number of frames, a time in seconds followed by s, or a timecode in the'
-    ' format HH:MM:SS or HH:MM:SS.nnn')
 @click.option(
     '--fade-bias', '-f', metavar='PERCENT',
     type=click.IntRange(-100, 100), default=0, show_default=True, help=
@@ -471,7 +486,7 @@ def detect_content_command(ctx, threshold, min_scene_len):
     type=click.IntRange(1, 128), default=8, show_default=True, help=
     'Number of rows in image to sum per iteration (can be tuned for performance in some cases).')
 @click.pass_context
-def detect_threshold_command(ctx, threshold, min_scene_len, fade_bias, add_last_scene,
+def detect_threshold_command(ctx, threshold, fade_bias, add_last_scene,
                              min_percent, block_size):
     """  Perform threshold detection algorithm on input video(s).
 
@@ -480,7 +495,9 @@ def detect_threshold_command(ctx, threshold, min_scene_len, fade_bias, add_last_
     detect-threshold --threshold 15
     """
 
-    logging.debug('Detecting threshold, parameters:\n'
+    min_scene_len = 0 if ctx.obj.drop_short_scenes else ctx.obj.min_scene_len
+
+    ctx.obj.logger.debug('Detecting threshold, parameters:\n'
                   '  threshold: %d, min-scene-len: %d, fade-bias: %d,\n'
                   '  add-last-scene: %s, min-percent: %d, block-size: %d',
                   threshold, min_scene_len, fade_bias,
@@ -489,14 +506,14 @@ def detect_threshold_command(ctx, threshold, min_scene_len, fade_bias, add_last_
     # Handle case where add_last_scene is not set and is None.
     add_last_scene = True if add_last_scene else False
 
-    min_scene_len = parse_timecode(ctx.obj, min_scene_len)
-
     # Convert min_percent and fade_bias from integer to floats (0.0-1.0 and -1.0-+1.0 respectively).
     min_percent /= 100.0
     fade_bias /= 100.0
     ctx.obj.add_detector(scenedetect.detectors.ThresholdDetector(
         threshold=threshold, min_scene_len=min_scene_len, fade_bias=fade_bias,
-        add_final_scene=add_last_scene, min_percent=min_percent, block_size=block_size))
+        add_final_scene=add_last_scene, block_size=block_size))
+
+
 
 
 @click.command('export-html', add_help_option=False)
@@ -521,7 +538,10 @@ def detect_threshold_command(ctx, threshold, min_scene_len, fade_bias, add_last_
 def export_html_command(ctx, filename, no_images, image_width, image_height):
     """ Exports scene list to a HTML file. Requires save-images by default."""
     if not ctx.obj.save_images and not no_images:
-        raise click.BadParameter("save-images isn't enabled")
+        ctx.obj.options_processed = False
+        raise click.BadArgumentUsage(
+            'The export-html command requires that the save-images command\n'
+            'is specified before it, unless --no-images is specified.')
     ctx.obj.export_html_command(filename, no_images, image_width, image_height)
     ctx.obj.export_html = True
 
@@ -547,13 +567,19 @@ def export_html_command(ctx, filename, no_images, image_width, image_height):
     '--quiet', '-q',
     is_flag=True, flag_value=True, help=
     'Suppresses output of the table printed by the list-scenes command.')
+@click.option(
+    '--skip-cuts', '-s',
+    is_flag=True, flag_value=True, help=
+    'Skips outputting the cutting list as the first row in the CSV file.'
+    ' Set this option if compliance with RFC 4810 is required.')
 @click.pass_context
-def list_scenes_command(ctx, output, filename, no_output_file, quiet):
+def list_scenes_command(ctx, output, filename, no_output_file, quiet, skip_cuts):
     """ Prints scene list and outputs to a CSV file. The default filename is
     $VIDEO_NAME-Scenes.csv. """
     if ctx.obj.list_scenes:
         duplicate_command(ctx, 'list-scenes')
-    ctx.obj.list_scenes_command(output, filename, no_output_file, quiet)
+    ctx.obj.list_scenes_command(
+        output, filename, no_output_file, quiet, skip_cuts)
     ctx.obj.list_scenes = True
 
 
@@ -613,8 +639,16 @@ def split_video_command(ctx, output, filename, high_quality, override_args, quie
                         rate_factor, preset):
     """Split input video(s) using ffmpeg or mkvmerge."""
     if ctx.obj.split_video:
-        logging.warning('split-video command is specified twice.')
+        duplicate_command(ctx, 'split-video')
     ctx.obj.check_input_open()
+    check_split_video_requirements(copy)
+
+    if contains_sequence_or_url(ctx.obj.video_manager.get_video_paths()):
+        ctx.obj.options_processed = False
+        error_str = 'The save-images command is incompatible with image sequences/URLs.'
+        ctx.obj.logger.error(error_str)
+        raise click.BadParameter(error_str, param_hint='save-images')
+
     ctx.obj.split_video = True
     ctx.obj.split_quiet = True if quiet else False
     ctx.obj.split_directory = output
@@ -622,9 +656,9 @@ def split_video_command(ctx, output, filename, high_quality, override_args, quie
     if copy:
         ctx.obj.split_mkvmerge = True
         if high_quality:
-            logging.warning('-hq/--high-quality flag ignored due to -c/--copy.')
+            ctx.obj.logger.warning('-hq/--high-quality flag ignored due to -c/--copy.')
         if override_args:
-            logging.warning('-f/--ffmpeg-args option ignored due to -c/--copy.')
+            ctx.obj.logger.warning('-f/--ffmpeg-args option ignored due to -c/--copy.')
     if not override_args:
         if rate_factor is None:
             rate_factor = 22 if not high_quality else 17
@@ -633,35 +667,13 @@ def split_video_command(ctx, output, filename, high_quality, override_args, quie
         override_args = ('-c:v libx264 -preset {PRESET} -crf {RATE_FACTOR} -c:a aac'.format(
             PRESET=preset, RATE_FACTOR=rate_factor))
     if not copy:
-        logging.info('FFmpeg codec args set: %s', override_args)
+        ctx.obj.logger.info('FFmpeg codec args set: %s', override_args)
     if filename:
-        logging.info('Video output file name format: %s', filename)
+        ctx.obj.logger.info('Video output file name format: %s', filename)
     if ctx.obj.split_directory is not None:
-        logging.info('Video output path set:  \n%s', ctx.obj.split_directory)
+        ctx.obj.logger.info('Video output path set:  \n%s', ctx.obj.split_directory)
     ctx.obj.split_args = override_args
 
-    mkvmerge_available = is_mkvmerge_available()
-    ffmpeg_available = is_ffmpeg_available()
-    if not (mkvmerge_available or ffmpeg_available) or (
-            (not mkvmerge_available and copy) or (not ffmpeg_available and not copy)):
-        split_tool = 'ffmpeg/mkvmerge'
-        if (not mkvmerge_available and copy):
-            split_tool = 'mkvmerge'
-        elif (not ffmpeg_available and not copy):
-            split_tool = 'ffmpeg'
-        error_strs = [
-            "{EXTERN_TOOL} is required for split-video{EXTRA_ARGS}.".format(
-                EXTERN_TOOL=split_tool, EXTRA_ARGS=' -c/--copy' if copy else ''),
-            "Install the above tool%s to enable video splitting support." % (
-                's' if split_tool.find('/') > 0 else '')]
-        if mkvmerge_available:
-            error_strs += [
-                'You can also specify `split-video -c/--copy` to use mkvmerge for splitting.']
-        error_str = '\n'.join(error_strs)
-        logging.debug(error_str)
-        ctx.obj.options_processed = False
-        raise click.BadParameter(error_str, param_hint='split-video')
-
 
 
 @click.command('save-images', add_help_option=False)
@@ -673,8 +685,8 @@ def split_video_command(ctx, output, filename, high_quality, override_args, quie
     '--filename', '-f', metavar='NAME', default='$VIDEO_NAME-Scene-$SCENE_NUMBER-$IMAGE_NUMBER',
     type=click.STRING, show_default=True, help=
     'Filename format, *without* extension, to use when saving image files. You can use the'
-    ' $VIDEO_NAME, $SCENE_NUMBER, and $IMAGE_NUMBER macros in the file name. Note that you'
-    ' may have to wrap the format in single quotes.')
+    ' $VIDEO_NAME, $SCENE_NUMBER, $IMAGE_NUMBER, and $FRAME_NUMBER macros in the file name.'
+    ' Note that you may have to wrap the format in single quotes.')
 @click.option(
     '--num-images', '-n', metavar='N', default=3,
     type=click.INT, help=
@@ -699,23 +711,46 @@ def split_video_command(ctx, output, filename, high_quality, override_args, quie
     is_flag=True, flag_value=True, help=
     'Set output format to PNG.')
 @click.option(
-    '--compression', '-c', metavar='C', default=None,
+    '--compression', '-c', metavar='C', default=3, show_default=False,
     type=click.IntRange(0, 9), help=
     'PNG compression rate, from 0-9. Higher values produce smaller files but result'
     ' in longer compression time. This setting does not affect image quality, only'
     ' file size. [default: 3]')
 @click.option(
-    '--image-frame-margin', metavar='N', default=0,
+    '-m', '--frame-margin', metavar='N', default=1, show_default=True,
     type=click.INT, help=
     'Number of frames to ignore at the beginning and end of scenes when saving images')
+@click.option(
+    '--scale', '-s', metavar='S', default=None, show_default=False,
+    type=click.FLOAT, help=
+    'Optional factor by which saved images are rescaled. A scaling factor of 1 would'
+    ' not result in rescaling. A value <1 results in a smaller saved image, while a'
+    ' value >1 results in an image larger than the original. This value is ignored if'
+    ' either the height, -h, or width, -w, values are specified.')
+@click.option(
+    '--height', '-h', metavar='H', default=None, show_default=False,
+    type=click.INT, help=
+    'Optional value for the height of the saved images. Specifying both the height'
+    ' and width, -w, will resize images to an exact size, regardless of aspect ratio.'
+    ' Specifying only height will rescale the image to that number of pixels in height'
+    ' while preserving the aspect ratio.')
+@click.option(
+    '--width', '-w', metavar='W', default=None, show_default=False,
+    type=click.INT, help=
+    'Optional value for the width of the saved images. Specifying both the width'
+    ' and height, -h, will resize images to an exact size, regardless of aspect ratio.'
+    ' Specifying only width will rescale the image to that number of pixels wide'
+    ' while preserving the aspect ratio.')
 @click.pass_context
 def save_images_command(ctx, output, filename, num_images, jpeg, webp, quality, png,
-                        compression, image_frame_margin):
+                        compression, frame_margin, scale, height, width):
     """ Create images for each detected scene. """
     if ctx.obj.save_images:
         duplicate_command(ctx, 'save-images')
+    if quality is None:
+        quality = 100 if webp else 95
     ctx.obj.save_images_command(num_images, output, filename, jpeg, webp, quality, png,
-                                compression, image_frame_margin)
+                                compression, frame_margin, scale, height, width)
 
 
 
@@ -738,18 +773,25 @@ def colors_command(ctx):
 
 
 
-# Info/Terminating Commands:
+# ----------------------------------------------------------------------
+# Commands Added To Help List
+# ----------------------------------------------------------------------
+
+# Info/Terminating Commands
 add_cli_command(scenedetect_cli, help_command)
-add_cli_command(scenedetect_cli, about_command)
 add_cli_command(scenedetect_cli, version_command)
+add_cli_command(scenedetect_cli, about_command)
 
-# Commands Added To Help List:
+# Input Commands
 add_cli_command(scenedetect_cli, time_command)
-add_cli_command(scenedetect_cli, detect_content_command)
-add_cli_command(scenedetect_cli, detect_threshold_command)
-add_cli_command(scenedetect_cli, list_scenes_command)
 
+# Output Commands
+add_cli_command(scenedetect_cli, export_html_command)
+add_cli_command(scenedetect_cli, list_scenes_command)
 add_cli_command(scenedetect_cli, save_images_command)
 add_cli_command(scenedetect_cli, split_video_command)
 
-add_cli_command(scenedetect_cli, export_html_command)
+# Detection Algorithms
+add_cli_command(scenedetect_cli, detect_content_command)
+add_cli_command(scenedetect_cli, detect_threshold_command)
+add_cli_command(scenedetect_cli, detect_adaptive_command)
