@@ -6,7 +6,7 @@
 #     [  Github: https://github.com/Breakthrough/PySceneDetect/  ]
 #     [  Documentation: http://pyscenedetect.readthedocs.org/    ]
 #
-# Copyright (C) 2014-2020 Brandon Castellano <http://www.bcastell.com>.
+# Copyright (C) 2014-2021 Brandon Castellano <http://www.bcastell.com>.
 #
 # PySceneDetect is licensed under the BSD 3-Clause License; see the included
 # LICENSE file, or visit one of the following pages for details:
@@ -36,26 +36,22 @@ from __future__ import print_function
 import logging
 import os
 import time
-import math
 from string import Template
-import numpy as np
 
 # Third-Party Library Imports
 import click
 import cv2
-from scenedetect.platform import tqdm
-from scenedetect.platform import get_and_create_path
 
 # PySceneDetect Library Imports
 import scenedetect.detectors
 
 from scenedetect.scene_manager import SceneManager
+from scenedetect.scene_manager import save_images
 from scenedetect.scene_manager import write_scene_list
 from scenedetect.scene_manager import write_scene_list_html
 
 from scenedetect.stats_manager import StatsManager
 from scenedetect.stats_manager import StatsFileCorrupt
-from scenedetect.stats_manager import StatsFileFramerateMismatch
 
 from scenedetect.video_manager import VideoManager
 from scenedetect.video_manager import VideoOpenFailure
@@ -70,11 +66,36 @@ from scenedetect.video_splitter import split_video_ffmpeg
 
 from scenedetect.platform import get_cv2_imwrite_params
 from scenedetect.platform import check_opencv_ffmpeg_dll
-
+from scenedetect.platform import get_and_create_path
 
 from scenedetect.frame_timecode import FrameTimecode
 
+
+def parse_timecode(cli_ctx, value):
+    # type: (CliContext, str) -> Union[FrameTimecode, None]
+    """ Parses a user input string expected to be a timecode, given a CLI context.
+
+    Returns:
+        (FrameTimecode) Timecode set to value with the CliContext VideoManager framerate.
+            If value is None, skips processing and returns None.
+
+    Raises:
+        click.BadParameter
+     """
+    cli_ctx.check_input_open()
+    if value is None:
+        return value
+    try:
+        timecode = FrameTimecode(
+            timecode=value, fps=cli_ctx.video_manager.get_framerate())
+        return timecode
+    except (ValueError, TypeError):
+        raise click.BadParameter(
+            'timecode must be in frames (1234), seconds (123.4s), or HH:MM:SS (00:02:03.400)')
+
+
 def get_plural(val_list):
+    # type: (List[any]) -> str
     """ Get Plural: Helper function to return 's' if a list has more than one (1)
     element, otherwise returns ''.
 
@@ -84,6 +105,49 @@ def get_plural(val_list):
     return 's' if len(val_list) > 1 else ''
 
 
+def contains_sequence_or_url(video_paths):
+    # type: (List[str]) -> bool
+    """ Checks if any of the video paths are a URL or image sequence.
+
+    Arguments:
+        video_paths: List of strings.
+
+    Returns: bool: True if any of the video_paths are a URL or image sequence,
+        False otherwise.
+    """
+
+    return any(['%' in video_path or '://' in video_path
+            for video_path in video_paths])
+
+
+def check_split_video_requirements(use_mkvmerge):
+    # type: (bool) -> None
+    """ Validates that the proper tool is available on the system to perform the split-video
+    command, which depends on if -c/--copy is set (to use mkvmerge) or not (to use ffmpeg).
+
+    Arguments:
+        use_mkvmerge: True if -c/--copy is set, False otherwise.
+
+    Raises: click.BadParameter if the proper video splitting tool cannot be found.
+    """
+
+    if (use_mkvmerge and not is_mkvmerge_available()) or not is_ffmpeg_available():
+        error_strs = [
+            "{EXTERN_TOOL} is required for split-video{EXTRA_ARGS}.".format(
+                EXTERN_TOOL='mkvmerge' if use_mkvmerge else 'ffmpeg',
+                EXTRA_ARGS=' -c/--copy' if use_mkvmerge else '')]
+        error_strs += ["Install one of the above tools to enable the split-video command."]
+        if not use_mkvmerge and is_mkvmerge_available():
+            error_strs += [
+                'You can also specify `-c/--copy` to use mkvmerge for splitting.']
+        elif use_mkvmerge and is_ffmpeg_available():
+            error_strs += [
+                'You can also omit `-c/--copy` to use ffmpeg for splitting.']
+        error_str = '\n'.join(error_strs)
+        raise click.BadParameter(error_str, param_hint='split-video')
+
+
+
 class CliContext(object):
     """ Context of the command-line interface passed between the various sub-commands.
 
@@ -108,6 +172,9 @@ class CliContext(object):
         self.output_directory = None            # -o/--output
         self.quiet_mode = False                 # -q/--quiet or -v/--verbosity quiet
         self.frame_skip = 0                     # -fs/--frame-skip
+        self.drop_short_scenes = False          # --drop-short-scenes
+        self.min_scene_len = None               # -m/--min-scene-len
+
         # Properties for save-images command.
         self.save_images = False                # save-images command
         self.image_extension = 'jpg'            # save-images -j/--jpeg, -w/--webp, -p/--png
@@ -115,13 +182,14 @@ class CliContext(object):
 
         self.image_param = None                 # save-images -q/--quality if -j/-w,
                                                 #   -c/--compression if -p
-
-
         self.image_name_format = (              # save-images -f/--name-format
             '$VIDEO_NAME-Scene-$SCENE_NUMBER-$IMAGE_NUMBER')
-        self.num_images = 2                     # save-images -n/--num-images
-        self.image_frame_margin = 0             # save-images --image-frame-margin
-        self.imwrite_params = get_cv2_imwrite_params()
+        self.num_images = 3                     # save-images -n/--num-images
+        self.frame_margin = 1                   # save-images -m/--frame-margin
+        self.scale = None                       # save-images -s/--scale
+        self.height = None                      # save-images -h/--height
+        self.width = None                       # save-images -w/--width
+
         # Properties for split-video command.
         self.split_video = False                # split-video command
         self.split_mkvmerge = False             # split-video -c/--copy
@@ -129,132 +197,35 @@ class CliContext(object):
         self.split_directory = None             # split-video -o/--output
         self.split_name_format = '$VIDEO_NAME-Scene-$SCENE_NUMBER'  # split-video -f/--filename
         self.split_quiet = False                # split-video -q/--quiet
+
         # Properties for list-scenes command.
         self.list_scenes = False                # list-scenes command
         self.print_scene_list = False           # list-scenes --quiet/-q
         self.scene_list_directory = None        # list-scenes -o/--output
         self.scene_list_name_format = None      # list-scenes -f/--filename
         self.scene_list_output = False          # list-scenes -n/--no-output
+        self.skip_cuts = False                  # list-scenes -s/--skip-cuts
 
+        # Properties for export-html command.
         self.export_html = False                # export-html command
         self.html_name_format = None            # export-html -f/--filename
         self.html_include_images = True         # export-html --no-images
-        self.image_filenames = None             # export-html used for embedding images
         self.image_width = None                 # export-html -w/--image-width
         self.image_height = None                # export-html -h/--image-height
 
+        # Logger for CLI output.
+        self.logger = logging.getLogger('pyscenedetect')
+
 
     def cleanup(self):
         # type: () -> None
         """ Cleanup: Releases all resources acquired by the CliContext (esp. the VideoManager). """
         try:
-            logging.debug('Cleaning up...\n\n')
+            self.logger.debug('Cleaning up...\n\n')
         finally:
             if self.video_manager is not None:
                 self.video_manager.release()
 
-    # TODO: Replace with scenedetect.scene_manager.save_images
-    def _generate_images(self, scene_list, video_name,
-                         image_name_template='$VIDEO_NAME-Scene-$SCENE_NUMBER-$IMAGE_NUMBER',
-                         output_dir=None):
-        # type: (List[Tuple[FrameTimecode, FrameTimecode]) -> None
-
-        if not scene_list:
-            return
-        if not self.options_processed:
-            return
-        if self.num_images <= 0:
-            raise ValueError()
-        self.check_input_open()
-
-        imwrite_param = []
-        if self.image_param is not None:
-            imwrite_param = [self.imwrite_params[self.image_extension], self.image_param]
-
-        # Reset video manager and downscale factor.
-        self.video_manager.release()
-        self.video_manager.reset()
-        self.video_manager.set_downscale_factor(1)
-        self.video_manager.start()
-
-        # Setup flags and init progress bar if available.
-        completed = True
-        logging.info('Generating output images (%d per scene)...', self.num_images)
-        progress_bar = None
-        if tqdm and not self.quiet_mode:
-            progress_bar = tqdm(
-                total=len(scene_list) * self.num_images, unit='images')
-
-        filename_template = Template(image_name_template)
-
-
-        scene_num_format = '%0'
-        scene_num_format += str(max(3, math.floor(math.log(len(scene_list), 10)) + 1)) + 'd'
-        image_num_format = '%0'
-        image_num_format += str(math.floor(math.log(self.num_images, 10)) + 2) + 'd'
-
-        timecode_list = dict()
-
-        fps = scene_list[0][0].framerate
-
-        timecode_list = [
-            [
-                FrameTimecode(int(f), fps=fps) for f in [
-                    # middle frames
-                    a[len(a)//2] if (0 < j < self.num_images-1) or self.num_images == 1
-
-                    # first frame
-                    else min(a[0] + self.image_frame_margin, a[-1]) if j == 0
-
-                    # last frame
-                    else max(a[-1] - self.image_frame_margin, a[0])
-
-                    # for each evenly-split array of frames in the scene list
-                    for j, a in enumerate(np.array_split(r, self.num_images))
-                ]
-            ]
-            for i, r in enumerate([
-                # pad ranges to number of images
-                r
-                if 1+r[-1]-r[0] >= self.num_images
-                else list(r) + [r[-1]] * (self.num_images - len(r))
-                # create range of frames in scene
-                for r in (
-                    range(start.get_frames(), end.get_frames())
-                    # for each scene in scene list
-                    for start, end in scene_list
-                    )
-            ])
-        ]
-
-        self.image_filenames = {i: [] for i in range(len(timecode_list))}
-
-        for i, scene_timecodes in enumerate(timecode_list):
-            for j, image_timecode in enumerate(scene_timecodes):
-                self.video_manager.seek(image_timecode)
-                self.video_manager.grab()
-                ret_val, frame_im = self.video_manager.retrieve()
-                if ret_val:
-                    file_path = '%s.%s' % (filename_template.safe_substitute(
-                        VIDEO_NAME=video_name,
-                        SCENE_NUMBER=scene_num_format % (i + 1),
-                        IMAGE_NUMBER=image_num_format % (j + 1),
-                        FRAME_NUMBER=image_timecode.get_frames()),
-                                           self.image_extension)
-                    self.image_filenames[i].append(file_path)
-                    cv2.imwrite(
-                        get_and_create_path(
-                            file_path,
-                            output_dir if output_dir is not None else self.output_directory),
-                        frame_im, imwrite_param)
-                else:
-                    completed = False
-                    break
-                if progress_bar:
-                    progress_bar.update(1)
-
-        if not completed:
-            logging.error('Could not generate all output images.')
 
     def _open_stats_file(self):
 
@@ -263,36 +234,23 @@ class CliContext(object):
 
         if self.stats_file_path is not None:
             if os.path.exists(self.stats_file_path):
-                logging.info('Loading frame metrics from stats file: %s',
+                self.logger.info('Loading frame metrics from stats file: %s',
                              os.path.basename(self.stats_file_path))
                 try:
                     with open(self.stats_file_path, 'rt') as stats_file:
-                        self.stats_manager.load_from_csv(stats_file, self.base_timecode)
+                        self.stats_manager.load_from_csv(stats_file)
                 except StatsFileCorrupt:
+                    error_info = (
+                        'Could not load frame metrics from stats file - file is either corrupt,'
+                        ' or not a valid PySceneDetect stats file. If the file exists, ensure that'
+                        ' it is a valid stats file CSV, otherwise delete it and run PySceneDetect'
+                        ' again to re-generate the stats file.')
                     error_strs = [
-                        'Could not load stats file.', 'Failed to parse stats file:',
-                        'Could not load frame metrics from stats file - file is corrupt or not a'
-                        ' valid PySceneDetect stats file. If the file exists, ensure that it is'
-                        ' a valid stats file CSV, otherwise delete it and run PySceneDetect again'
-                        ' to re-generate the stats file.']
-                    logging.error('\n'.join(error_strs))
+                        'Could not load stats file.', 'Failed to parse stats file:', error_info ]
+                    self.logger.error('\n'.join(error_strs))
                     raise click.BadParameter(
                         '\n  Could not load given stats file, see above output for details.',
                         param_hint='input stats file')
-                except StatsFileFramerateMismatch as ex:
-                    error_strs = [
-                        'could not load stats file.', 'Failed to parse stats file:',
-                        'Framerate differs between stats file (%.2f FPS) and input'
-                        ' video%s (%.2f FPS)' % (
-                            ex.stats_file_fps,
-                            's' if self.video_manager.get_num_videos() > 1 else '',
-                            ex.base_timecode_fps),
-                        'Ensure the correct stats file path was given, or delete and re-generate'
-                        ' the stats file.']
-                    logging.error('\n'.join(error_strs))
-                    raise click.BadParameter(
-                        'framerate differs between given stats file and input video(s).',
-                        param_hint='input stats file')
 
 
     def process_input(self):
@@ -301,74 +259,89 @@ class CliContext(object):
 
         Run after all command line options/sub-commands have been parsed.
         """
-        logging.debug('Processing input...')
+        self.logger.debug('Processing input...')
         if not self.options_processed:
-            logging.debug('Skipping processing, CLI options were not parsed successfully.')
+            self.logger.debug('Skipping processing, CLI options were not parsed successfully.')
             return
         self.check_input_open()
         assert self.scene_manager.get_num_detectors() >= 0
         if self.scene_manager.get_num_detectors() == 0:
-            logging.error(
+            self.logger.error(
                 'No scene detectors specified (detect-content, detect-threshold, etc...),\n'
                 ' or failed to process all command line arguments.')
             return
 
+        # Display a warning if the video codec type seems unsupported (#86).
+        if int(abs(self.video_manager.get(cv2.CAP_PROP_FOURCC))) == 0:
+            self.logger.error(
+                'Video codec detection failed, output may be incorrect.\nThis could be caused'
+                ' by using an outdated version of OpenCV, or using codecs that currently are'
+                ' not well supported (e.g. VP9).\n'
+                'As a workaround, consider re-encoding the source material before processing.\n'
+                'For details, see https://github.com/Breakthrough/PySceneDetect/issues/86')
+
         # Handle scene detection commands (detect-content, detect-threshold, etc...).
         self.video_manager.start()
-        base_timecode = self.video_manager.get_base_timecode()
 
         start_time = time.time()
-        logging.info('Detecting scenes...')
+        self.logger.info('Detecting scenes...')
 
         num_frames = self.scene_manager.detect_scenes(
             frame_source=self.video_manager, frame_skip=self.frame_skip,
             show_progress=not self.quiet_mode)
 
         # Handle case where video fails with multiple audio tracks (#179).
-        # TODO: Is there a fix for this? See #179.
+        # TODO: Using a different video backend as per #213 may also resolve this issue,
+        # as well as numerous other timing related issues.
         if num_frames <= 0:
-            logging.critical('\n'.join([
+            self.logger.critical(
                 'Failed to read any frames from video file. This could be caused'
                 ' by the video having multiple audio tracks. If so, please try'
-                ' removing the audio tracks or muxing to mkv via:'
-                '      ffmpeg -i input.mp4 -c copy -an output.mp4'
-                'or:'
-                '      mkvmerge -o output.mkv input.mp4'
-                ' For details, see https://pyscenedetect.readthedocs.io/en/latest/faq/']))
+                ' removing the audio tracks or muxing to mkv via:\n'
+                '      ffmpeg -i input.mp4 -c copy -an output.mp4\n'
+                'or:\n'
+                '      mkvmerge -o output.mkv input.mp4\n'
+                'For details, see https://pyscenedetect.readthedocs.io/en/latest/faq/')
             return
 
         duration = time.time() - start_time
-        logging.info('Processed %d frames in %.1f seconds (average %.2f FPS).',
+        self.logger.info('Processed %d frames in %.1f seconds (average %.2f FPS).',
                      num_frames, duration, float(num_frames)/duration)
 
         # Handle -s/--statsfile option.
         if self.stats_file_path is not None:
             if self.stats_manager.is_save_required():
                 with open(self.stats_file_path, 'wt') as stats_file:
-                    logging.info('Saving frame metrics to stats file: %s',
+                    self.logger.info('Saving frame metrics to stats file: %s',
                                  os.path.basename(self.stats_file_path))
+                    base_timecode = self.video_manager.get_base_timecode()
                     self.stats_manager.save_to_csv(
                         stats_file, base_timecode)
             else:
-                logging.debug('No frame metrics updated, skipping update of the stats file.')
+                self.logger.debug('No frame metrics updated, skipping update of the stats file.')
 
         # Get list of detected cuts and scenes from the SceneManager to generate the required output
         # files with based on the given commands (list-scenes, split-video, save-images, etc...).
-        cut_list = self.scene_manager.get_cut_list(base_timecode)
-        scene_list = self.scene_manager.get_scene_list(base_timecode)
-        video_paths = self.video_manager.get_video_paths()
-        video_name = os.path.basename(video_paths[0])
-        if video_name.rfind('.') >= 0:
-            video_name = video_name[:video_name.rfind('.')]
+        cut_list = self.scene_manager.get_cut_list()
+        scene_list = self.scene_manager.get_scene_list()
 
-        # Ensure we don't divide by zero.
-        if scene_list:
-            logging.info('Detected %d scenes, average shot length %.1f seconds.',
+        # Handle --drop-short-scenes.
+        if self.drop_short_scenes and self.min_scene_len > 0:
+            scene_list = [
+                s for s in scene_list
+                if (s[1] - s[0]) >= self.min_scene_len
+            ]
+
+        video_paths = self.video_manager.get_video_paths()
+        video_name = self.video_manager.get_video_name()
+
+        if scene_list:  # Ensure we don't divide by zero.
+            self.logger.info('Detected %d scenes, average shot length %.1f seconds.',
                          len(scene_list),
                          sum([(end_time - start_time).get_seconds()
                               for start_time, end_time in scene_list]) / float(len(scene_list)))
         else:
-            logging.info('No scenes detected.')
+            self.logger.info('No scenes detected.')
 
         # Handle list-scenes command.
         if self.scene_list_output:
@@ -380,12 +353,16 @@ class CliContext(object):
                 scene_list_filename,
                 self.scene_list_directory if self.scene_list_directory is not None
                 else self.output_directory)
-            logging.info('Writing scene list to CSV file:\n  %s', scene_list_path)
+            self.logger.info('Writing scene list to CSV file:\n  %s', scene_list_path)
             with open(scene_list_path, 'wt') as scene_list_file:
-                write_scene_list(scene_list_file, scene_list, cut_list)
-        # Handle `list-scenes`.
+                write_scene_list(
+                    output_csv_file=scene_list_file,
+                    scene_list=scene_list,
+                    include_cut_list=not self.skip_cuts,
+                    cut_list=cut_list)
+
         if self.print_scene_list:
-            logging.info("""Scene List:
+            self.logger.info("""Scene List:
 -----------------------------------------------------------------------
  | Scene # | Start Frame |  Start Time  |  End Frame  |   End Time   |
 -----------------------------------------------------------------------
@@ -398,16 +375,30 @@ class CliContext(object):
         end_time.get_frames(), end_time.get_timecode())
      for i, (start_time, end_time) in enumerate(scene_list)]))
 
-
         if cut_list:
-            logging.info('Comma-separated timecode list:\n  %s',
+            self.logger.info('Comma-separated timecode list:\n  %s',
                          ','.join([cut.get_timecode() for cut in cut_list]))
 
         # Handle save-images command.
+
         if self.save_images:
-            self._generate_images(scene_list=scene_list, video_name=video_name,
-                                  image_name_template=self.image_name_format,
-                                  output_dir=self.image_directory)
+            image_output_dir = self.output_directory
+            if self.image_directory is not None:
+                image_output_dir = self.image_directory
+
+            image_filenames = save_images(
+                scene_list=scene_list,
+                video_manager=self.video_manager,
+                num_images=self.num_images,
+                frame_margin=self.frame_margin,
+                image_extension=self.image_extension,
+                encoder_param=self.image_param,
+                image_name_template=self.image_name_format,
+                output_dir=image_output_dir,
+                show_progress=not self.quiet_mode,
+                scale=self.scale,
+                height=self.height,
+                width=self.width)
 
         # Handle export-html command.
         if self.export_html:
@@ -419,60 +410,42 @@ class CliContext(object):
                 html_filename,
                 self.image_directory if self.image_directory is not None
                 else self.output_directory)
-            logging.info('Exporting to html file:\n %s:', html_path)
+            self.logger.info('Exporting to html file:\n %s:', html_path)
             if not self.html_include_images:
-                self.image_filenames = None
+                image_filenames = None
             write_scene_list_html(html_path, scene_list, cut_list,
-                                  image_filenames=self.image_filenames,
+                                  image_filenames=image_filenames,
                                   image_width=self.image_width,
                                   image_height=self.image_height)
 
         # Handle split-video command.
         if self.split_video:
+            output_path_template = self.split_name_format
             # Add proper extension to filename template if required.
-            dot_pos = self.split_name_format.rfind('.')
-            if self.split_mkvmerge and not self.split_name_format.endswith('.mkv'):
-                self.split_name_format += '.mkv'
-            # Don't add if we find an extension between 2 and 4 characters
-            elif not (dot_pos >= 0) or (
-                    dot_pos >= 0 and not
-                    ((len(self.split_name_format) - (dot_pos+1) <= 4 >= 2))):
-                self.split_name_format += '.mp4'
-
-            output_file_prefix = get_and_create_path(
-                self.split_name_format,
+            dot_pos = output_path_template.rfind('.')
+            extension_length = 0 if dot_pos < 0 else len(output_path_template) - (dot_pos + 1)
+            # If using mkvmerge, force extension to .mkv.
+            if self.split_mkvmerge and not output_path_template.endswith('.mkv'):
+                output_path_template += '.mkv'
+            # Otherwise, if using ffmpeg, only add an extension if one doesn't exist.
+            elif not 2 <= extension_length <= 4:
+                output_path_template += '.mp4'
+            output_path_template = get_and_create_path(
+                output_path_template,
                 self.split_directory if self.split_directory is not None
                 else self.output_directory)
-            mkvmerge_available = is_mkvmerge_available()
-            ffmpeg_available = is_ffmpeg_available()
-            if mkvmerge_available and (self.split_mkvmerge or not ffmpeg_available):
-                if not self.split_mkvmerge:
-                    logging.warning(
-                        'ffmpeg not found, falling back to fast copy mode (split-video -c/--copy).')
-                split_video_mkvmerge(video_paths, scene_list, output_file_prefix, video_name,
+            # Ensure the appropriate tool is available before handling split-video.
+            check_split_video_requirements(self.split_mkvmerge)
+            if self.split_mkvmerge:
+                split_video_mkvmerge(video_paths, scene_list, output_path_template, video_name,
                                      suppress_output=self.quiet_mode or self.split_quiet)
-            elif ffmpeg_available:
-                if self.split_mkvmerge:
-                    logging.warning('mkvmerge not found, falling back to normal splitting'
-                                    ' mode (split-video).')
-                split_video_ffmpeg(video_paths, scene_list, output_file_prefix,
+            else:
+                split_video_ffmpeg(video_paths, scene_list, output_path_template,
                                    video_name, arg_override=self.split_args,
                                    hide_progress=self.quiet_mode,
                                    suppress_output=self.quiet_mode or self.split_quiet)
-            else:
-                if not (mkvmerge_available or ffmpeg_available):
-                    error_strs = ["ffmpeg/mkvmerge is required for split-video [-c/--copy]."]
-                else:
-                    error_strs = [
-                        "{EXTERN_TOOL} is required for split-video{EXTRA_ARGS}.".format(
-                            EXTERN_TOOL='mkvmerge' if self.split_mkvmerge else 'ffmpeg',
-                            EXTRA_ARGS=' -c/--copy' if self.split_mkvmerge else '')]
-                error_strs += ["Install one of the above tools to enable the split-video command."]
-                error_str = '\n'.join(error_strs)
-                logging.debug(error_str)
-                raise click.BadParameter(error_str, param_hint='split-video')
             if scene_list:
-                logging.info('Video splitting completed, individual scenes written to disk.')
+                self.logger.info('Video splitting completed, individual scenes written to disk.')
 
 
 
@@ -489,7 +462,7 @@ class CliContext(object):
             error_strs = ["No input video(s) specified.",
                           "Make sure '--input VIDEO' is specified at the start of the command."]
             error_str = '\n'.join(error_strs)
-            logging.debug(error_str)
+            self.logger.debug(error_str)
             raise click.BadParameter(error_str, param_hint='input video')
 
 
@@ -510,11 +483,11 @@ class CliContext(object):
 
         self.base_timecode = None
 
-        logging.debug('Initializing VideoManager.')
+        self.logger.debug('Initializing VideoManager.')
         video_manager_initialized = False
         try:
             self.video_manager = VideoManager(
-                video_files=input_list, framerate=framerate, logger=logging)
+                video_files=input_list, framerate=framerate, logger=self.logger)
             video_manager_initialized = True
             self.base_timecode = self.video_manager.get_base_timecode()
             self.video_manager.set_downscale_factor(downscale)
@@ -529,7 +502,7 @@ class CliContext(object):
                     'Error: OpenCV dependency %s not found.' % dll_name,
                     'Ensure that you installed the Python OpenCV module, and that the',
                     '%s file can be found to enable video support.' % dll_name]
-            logging.debug('\n'.join(error_strs[1:]))
+            self.logger.debug('\n'.join(error_strs[1:]))
             if not dll_okay:
                 click.echo(click.style(
                     '\nOpenCV dependency missing, video input/decoding not available.\n', fg='red'))
@@ -538,7 +511,7 @@ class CliContext(object):
             error_strs = ['could not get framerate from video(s)',
                           'Failed to obtain framerate for video file %s.' % ex.file_name]
             error_strs.append('Specify framerate manually with the -f / --framerate option.')
-            logging.debug('\n'.join(error_strs))
+            self.logger.debug('\n'.join(error_strs))
             raise click.BadParameter('\n'.join(error_strs), param_hint='input video')
         except VideoParameterMismatch as ex:
             error_strs = ['video parameters do not match.', 'List of mismatched parameters:']
@@ -554,16 +527,17 @@ class CliContext(object):
             error_strs.append(
                 'Multiple videos may only be specified if they have the same framerate and'
                 ' resolution. -f / --framerate may be specified to override the framerate.')
-            logging.debug('\n'.join(error_strs))
+            self.logger.debug('\n'.join(error_strs))
             raise click.BadParameter('\n'.join(error_strs), param_hint='input videos')
         except InvalidDownscaleFactor as ex:
             error_strs = ['Downscale value is not > 0.', str(ex)]
-            logging.debug('\n'.join(error_strs))
+            self.logger.debug('\n'.join(error_strs))
             raise click.BadParameter('\n'.join(error_strs), param_hint='downscale factor')
         return video_manager_initialized
 
 
-    def parse_options(self, input_list, framerate, stats_file, downscale, frame_skip):
+    def parse_options(self, input_list, framerate, stats_file, downscale, frame_skip,
+                      min_scene_len, drop_short_scenes):
         # type: (List[str], float, str, int, int) -> None
         """ Parse Options: Parses all global options/arguments passed to the main
         scenedetect command, before other sub-commands (e.g. this function processes
@@ -578,7 +552,7 @@ class CliContext(object):
         if not input_list:
             return
 
-        logging.debug('Parsing program options.')
+        self.logger.debug('Parsing program options.')
 
         self.frame_skip = frame_skip
 
@@ -588,9 +562,9 @@ class CliContext(object):
         # Ensure VideoManager is initialized, and open StatsManager if --stats is specified.
         if not video_manager_initialized:
             self.video_manager = None
-            logging.info('VideoManager not initialized.')
+            self.logger.info('VideoManager not initialized.')
         else:
-            logging.debug('VideoManager initialized.')
+            self.logger.debug('VideoManager initialized.')
             self.stats_file_path = get_and_create_path(stats_file, self.output_directory)
             if self.stats_file_path is not None:
                 self.check_input_open()
@@ -599,6 +573,9 @@ class CliContext(object):
         # Init SceneManager.
         self.scene_manager = SceneManager(self.stats_manager)
 
+        self.drop_short_scenes = drop_short_scenes
+        self.min_scene_len = parse_timecode(self, min_scene_len)
+
         self.options_processed = True
 
 
@@ -611,7 +588,7 @@ class CliContext(object):
         Raises:
             click.BadParameter, VideoDecodingInProgress
         """
-        logging.debug('Setting video time:\n    start: %s, duration: %s, end: %s',
+        self.logger.debug('Setting video time:\n    start: %s, duration: %s, end: %s',
                       start, duration, end)
 
         self.check_input_open()
@@ -627,7 +604,8 @@ class CliContext(object):
             self.start_frame = start.get_frames()
 
 
-    def list_scenes_command(self, output_path, filename_format, no_output_mode, quiet_mode):
+    def list_scenes_command(self, output_path, filename_format, no_output_mode,
+                            quiet_mode, skip_cuts):
         # type: (str, str, bool, bool) -> None
         """ List Scenes Command: Parses all options/arguments passed to the list-scenes command,
         or with respect to the CLI, this function processes [list-scenes options] when calling:
@@ -642,10 +620,11 @@ class CliContext(object):
         self.scene_list_directory = output_path
         self.scene_list_name_format = filename_format
         if self.scene_list_name_format is not None and not no_output_mode:
-            logging.info('Scene list CSV file name format:\n  %s', self.scene_list_name_format)
+            self.logger.info('Scene list CSV file name format:\n  %s', self.scene_list_name_format)
         self.scene_list_output = False if no_output_mode else True
         if self.scene_list_directory is not None:
-            logging.info('Scene list output directory set:\n  %s', self.scene_list_directory)
+            self.logger.info('Scene list output directory set:\n  %s', self.scene_list_directory)
+        self.skip_cuts = skip_cuts
 
 
     def export_html_command(self, filename, no_images, image_width, image_height):
@@ -661,15 +640,15 @@ class CliContext(object):
 
         self.html_name_format = filename
         if self.html_name_format is not None:
-            logging.info('Scene list html file name format:\n %s', self.html_name_format)
+            self.logger.info('Scene list html file name format:\n %s', self.html_name_format)
         self.html_include_images = False if no_images else True
         self.image_width = image_width
         self.image_height = image_height
 
 
     def save_images_command(self, num_images, output, name_format, jpeg, webp, quality,
-                            png, compression, image_frame_margin):
-        # type: (int, str, str, bool, bool, int, bool, int) -> None
+                            png, compression, frame_margin, scale, height, width):
+        # type: (int, str, str, bool, bool, int, bool, int, float, int, int) -> None
         """ Save Images Command: Parses all options/arguments passed to the save-images command,
         or with respect to the CLI, this function processes [save-images options] when calling:
         scenedetect [global options] save-images [save-images options] [other commands...].
@@ -679,7 +658,13 @@ class CliContext(object):
         """
         self.check_input_open()
 
-        num_flags = sum([True if flag else False for flag in [jpeg, webp, png]])
+        if contains_sequence_or_url(self.video_manager.get_video_paths()):
+            self.options_processed = False
+            error_str = '\nThe save-images command is incompatible with image sequences/URLs.'
+            self.logger.error(error_str)
+            raise click.BadParameter(error_str, param_hint='save-images')
+
+        num_flags = sum([1 if flag else 0 for flag in [jpeg, webp, png]])
         if num_flags <= 1:
 
             # Ensure the format exists.
@@ -688,13 +673,14 @@ class CliContext(object):
                 extension = 'png'
             elif webp:
                 extension = 'webp'
-            if not extension in self.imwrite_params or self.imwrite_params[extension] is None:
+            valid_params = get_cv2_imwrite_params()
+            if not extension in valid_params or valid_params[extension] is None:
                 error_strs = [
                     'Image encoder type %s not supported.' % extension.upper(),
                     'The specified encoder type could not be found in the current OpenCV module.',
                     'To enable this output format, please update the installed version of OpenCV.',
                     'If you build OpenCV, ensure the the proper dependencies are enabled. ']
-                logging.debug('\n'.join(error_strs))
+                self.logger.debug('\n'.join(error_strs))
                 raise click.BadParameter('\n'.join(error_strs), param_hint='save-images')
 
             self.save_images = True
@@ -703,19 +689,22 @@ class CliContext(object):
             self.image_param = compression if png else quality
             self.image_name_format = name_format
             self.num_images = num_images
-            self.image_frame_margin = image_frame_margin
+            self.frame_margin = frame_margin
+            self.scale = scale
+            self.height = height
+            self.width = width
 
             image_type = 'JPEG' if self.image_extension == 'jpg' else self.image_extension.upper()
             image_param_type = ''
             if self.image_param:
                 image_param_type = 'Compression' if image_type == 'PNG' else 'Quality'
                 image_param_type = ' [%s: %d]' % (image_param_type, self.image_param)
-            logging.info('Image output format set: %s%s', image_type, image_param_type)
+            self.logger.info('Image output format set: %s%s', image_type, image_param_type)
             if self.image_directory is not None:
-                logging.info('Image output directory set:\n  %s',
+                self.logger.info('Image output directory set:\n  %s',
                              os.path.abspath(self.image_directory))
         else:
             self.options_processed = False
-            logging.error('Multiple image type flags set for save-images command.')
+            self.logger.error('Multiple image type flags set for save-images command.')
             raise click.BadParameter(
                 'Only one image type (JPG/PNG/WEBP) can be specified.', param_hint='save-images')
